@@ -3,13 +3,13 @@
 ;; Copyright (C) 1995-1998 Frederic Lepied <Frederic.Lepied@sugix.frmug.org>
 ;;
 ;; Author: Frederic Lepied <Frederic.Lepied@sugix.frmug.org>
-;; Version: $Id: cvs.el,v 1.31 1998/09/07 12:00:17 fred Exp $
+;; Version: $Id: cvs.el,v 1.2 2001/12/20 14:37:27 adrian Exp $
 ;; Keywords: cvs ediff mode-line
 ;;
 ;; LCD Archive Entry:
 ;; cvs|Frederic Lepied|Frederic.Lepied@sugix.frmug.org|
 ;; Light cvs support for emacs (ediff + msb + dired + mode line).|
-;; $Date: 1998/09/07 12:00:17 $|$Revision: 1.31 $|~/modes/cvs.el.gz|
+;; $Date: 2001/12/20 14:37:27 $|$Revision: 1.2 $|~/modes/cvs.el.gz|
 ;;
 ;; This program is free  software;  you can redistribute   it and/or modify  it
 ;; under the terms of  the GNU General Public  License as published by the Free
@@ -51,7 +51,7 @@
 (require 'easymenu)
 
 ;;=============================================================================
-(defconst cvs:version "$Id: cvs.el,v 1.31 1998/09/07 12:00:17 fred Exp $"
+(defconst cvs:version "$Id: cvs.el,v 1.2 2001/12/20 14:37:27 adrian Exp $"
   "Version number of cvs.el. To communicate with bug report")
 
 ;;=============================================================================
@@ -107,20 +107,24 @@
 
 ;;=============================================================================
 (defvar cvs-shell-command
-  (if (memq system-type '(ms-dos emx))
+  (if (memq system-type '(ms-dos emx windows-nt))
       shell-file-name
     "/bin/sh")
-  "Name of the shell used in cvs commands.")
+  "Name of the shell used in cvs commands.
+It is initialized from `shell-file-name' on most systems.
+NB: Not all shells may be adequate as `shell-file-name'.")
 
 ;;=============================================================================
 (defvar cvs-shell-command-option
-  (cond ((eq system-type 'ms-dos)
-         (if (boundp 'shell-command-option)
-             shell-command-option
+  (cond ((memq system-type '(ms-dos windows-nt))
+         (if (boundp 'shell-command-switch)
+             shell-command-switch
            "/c"))
         (t                              ;Unix & EMX (Emacs 19 port to OS/2)
          "-c"))
-  "Shell argument indicating that next argument is the command.")
+  "Shell argument indicating that next argument is the command.
+It is initialized from `shell-command-switch' on most systems.
+NB: Not all shells may have an adequate `shell-command-switch'.")
 
 ;;=============================================================================
 (defvar cvs-file-option "-F"
@@ -736,8 +740,9 @@ If the file is read-only, runs the edit command else runs the unedit command."
 	  (message "Converting symbolic name to numeric (please wait) ...")
 	  (buffer-disable-undo tmpbuf)
 	  (erase-buffer tmpbuf)
-	  (call-process "/bin/sh" nil tmpbuf nil
-			"-c" (format cmd-format-args name file))
+	  (call-process cvs-shell-command nil tmpbuf nil
+                        cvs-shell-command-option
+                        (format cmd-format-args name file))
 	  (set-buffer tmpbuf)
 	  (goto-char (point-min))
 	  (if (not (re-search-forward "^VERS:[ \t]*\\([.0-9][.[0-9]*\\)[ \t]*$"
