@@ -698,6 +698,10 @@ Returns nil if line starts inside a string, t if in a comment."
     (end-of-icon-defun)
     (forward-char -1)))
 
+;; XEmacs addition
+;;;###autoload
+(add-to-list 'auto-mode-alist '("\\.icn\\'" . icon-mode))
+
 (provide 'icon)
 
 ;;; icon.el ends here
