@@ -1,41 +1,35 @@
-;; tcl.el --- Tcl code editing commands for Emacs
+;;; tcl.el --- Tcl code editing commands for Emacs
 
-;; Copyright (C) 1994 Free Software Foundation, Inc.
+;; Copyright (C) 1994, 1998, 1999, 2000, 2001  Free Software Foundation, Inc.
 
-;; Maintainer: Tom Tromey <tromey@busco.lanl.gov>
+;; Maintainer: FSF
 ;; Author: Tom Tromey <tromey@busco.lanl.gov>
 ;;    Chris Lindblad <cjl@lcs.mit.edu>
-;; Keywords: languages
-;; Version: 1.50
+;; Keywords: languages tcl modes
+;; Version: Revision: 1.64.4.1
 
 ;; This file is part of XEmacs.
 
-;; XEmacs is free software; you can redistribute it and/or modify it
-;; under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 1, or (at your option)
+;; XEmacs is free software; you can redistribute it and/or modify
+;; it under the terms of the GNU General Public License as published by
+;; the Free Software Foundation; either version 2, or (at your option)
 ;; any later version.
 
-;; XEmacs is distributed in the hope that it will be useful, but
-;; WITHOUT ANY WARRANTY; without even the implied warranty of
-;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-;; General Public License for more details.
+;; XEmacs is distributed in the hope that it will be useful,
+;; but WITHOUT ANY WARRANTY; without even the implied warranty of
+;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with XEmacs; see the file COPYING.  If not, write to the Free
-;; Software Foundation, Inc. 59 Temple Place - Suite 330,
+;; along with XEmacs; see the file COPYING.  If not, write to the
+;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 ;; Boston, MA 02111-1307, USA.
 
-;;; Synched up with: Not synched
+;;; Synched up with: Mostly with GNU Emacs 21.2.  The differences are a
+;;; couple of defcustom<->defvars, we use `tcl-imenu-generic-expression'
+;;; instead of `tcl-imenu-create-index-function' and some typo fixes.
 
-;;; Commentary:
-
-;; HOW TO INSTALL:
-;; Put the following forms in your .emacs to enable autoloading of Tcl
-;; mode, and auto-recognition of ".tcl" files.
-;;
-;;   (autoload 'tcl-mode "tcl" "Tcl mode." t)
-;;   (autoload 'inferior-tcl "tcl" "Run inferior Tcl process." t)
-;;   (setq auto-mode-alist (append '(("\\.tcl$" . tcl-mode)) auto-mode-alist))
+;; BEFORE USE:
 ;;
 ;; If you plan to use the interface to the TclX help files, you must
 ;; set the variable tcl-help-directory-list to point to the topmost
@@ -43,18 +37,7 @@
 ;;
 ;;   (setq tcl-help-directory-list '("/usr/local/lib/tclx/help"))
 ;;
-;; Also you will want to add the following to your .emacs:
-;;
-;;   (autoload 'tcl-help-on-word "tcl" "Help on Tcl commands" t)
-;;
-;; FYI a *very* useful thing to do is nroff all the Tk man pages and
-;; put them in a subdir of the help system.
-;;
-
-;; LCD Archive Entry:
-;; tcl|Tom Tromey|tromey@busco.lanl.gov|
-;; Major mode for editing Tcl|
-;; 1996/03/23 05:14:50|1.50|~/modes/tcl.el.Z|
+;;; Commentary:
 
 ;; CUSTOMIZATION NOTES:
 ;; * tcl-proc-list can be used to customize a list of things that
@@ -66,226 +49,7 @@
 ;; * tcl-type-alist can be used to minimally customize indentation
 ;; according to context.
 
-;; Change log:
-;; tcl.el,v
-;; Revision 1.50  1996/03/23  05:14:50  tromey
-;; (tcl-using-emacs-19): Work with XEmacs 20.0.  From Ben Wing.
-;;
-;; Revision 1.49  1995/12/07  18:27:47  tromey
-;; (add-log-tcl-defun): Don't use tcl-beginning-of-defun; just go to end
-;; of line before searching.
-;;
-;; Revision 1.48  1995/12/07  18:18:21  tromey
-;; (add-log-tcl-defun): Now uses tcl-beginning-of-defun.
-;;
-;; Revision 1.47  1995/08/22  17:49:45  tromey
-;; (tcl-hilit): New function from "Chris Alfeld" <calfeld@math.utah.edu>
-;; (tcl-mode): Call it
-;;
-;; Revision 1.46  1995/08/07  16:02:01  tromey
-;; (tcl-do-auto-fill): Only fill past fill-column; for 19.29.
-;; (tcl-auto-fill-mode): Use force-mode-line-update.
-;;
-;; Revision 1.45  1995/07/23  23:51:25  tromey
-;; (tcl-word-no-props): New function.
-;; (tcl-figure-type): Use it.
-;; (tcl-current-word): Ditto.
-;;
-;; Revision 1.44  1995/07/23  20:26:47  tromey
-;; Doc fixes.
-;;
-;; Revision 1.43  1995/07/17  19:59:49  tromey
-;; (inferior-tcl-mode): Use modeline-process if it exists.
-;;
-;; Revision 1.42  1995/07/17  19:55:25  tromey
-;; XEmacs currently must use tcl-internal-end-of-defun
-;;
-;; Revision 1.41  1995/07/14  21:54:56  tromey
-;; Changes to make menus work in XEmacs.
-;; From Mike Scheidler <c23mts@kocrsv01.delcoelect.com>
-;;
-;; Revision 1.40  1995/07/11  03:13:15  tromey
-;; (tcl-mode): Customize for new dabbrev.
-;;
-;; Revision 1.39  1995/07/09  21:58:03  tromey
-;; (tcl-do-fill-paragraph): New function.
-;; (tcl-mode): Set up for paragraph filling.
-;;
-;; Revision 1.38  1995/07/09  21:30:32  tromey
-;; (tcl-mode): Fixes to 19.29 paragraph variables.
-;;
-;; Revision 1.37  1995/07/09  18:52:16  tromey
-;; (tcl-do-auto-fill): Set fill-prefix.
-;;
-;; Revision 1.36  1995/07/09  01:07:57  tromey
-;; (tcl-imenu-create-index-function): Work with imenu from Emacs 19.29
-;;
-;; Revision 1.35  1995/06/27  20:12:00  tromey
-;; (tcl-type-alist): More itcl changes.
-;;
-;; Revision 1.34  1995/06/27  20:06:05  tromey
-;; More changes for itcl.
-;; Bug fixes for Emacs 19.29.
-;;
-;; Revision 1.33  1995/06/27  20:01:29  tromey
-;; (tcl-set-proc-regexp): Allow leading spaces.
-;; (tcl-proc-list): Changes for itcl.
-;; (tcl-typeword-list): Ditto.
-;; (tcl-keyword-list): Ditto.
-;;
-;; Revision 1.32  1995/05/11  22:12:49  tromey
-;; (tcl-type-alist): Include entry for "proc".
-;;
-;; Revision 1.31  1995/05/10  23:38:12  tromey
-;; (tcl-add-fsf-menu): Use make-lucid-menu-keymap, not
-;; "make-xemacs-menu-keymap".
-;;
-;; Revision 1.30  1995/05/10  18:22:21  tromey
-;; Bug fix in menu code for XEmacs.
-;;
-;; Revision 1.29  1995/05/09  21:36:53  tromey
-;; Changed "Lucid Emacs" to "XEmacs".
-;; Tcl's popup menu now added to existing one, courtesy
-;; dfarmer@evolving.com (Doug Farmer)
-;;
-;; Revision 1.28  1995/04/08  19:52:50  tromey
-;; (tcl-outline-level): New function
-;; (tcl-mode): Added outline-handling stuff.
-;; From Jesper Pedersen <blackie@imada.ou.dk>
-;;
-;; Revision 1.27  1994/10/11  02:01:27  tromey
-;; (tcl-mode): imenu-create-index-function made buffer local.
-;;
-;; Revision 1.26  1994/09/01  18:06:24  tromey
-;; Added filename completion in inferior tcl mode
-;;
-;; Revision 1.25  1994/08/22  15:56:24  tromey
-;; tcl-load-file default to current buffer.
-;;
-;; Revision 1.24  1994/08/21  20:33:05  tromey
-;; Fixed bug in tcl-guess-application.
-;;
-;; Revision 1.23  1994/08/21  03:54:45  tromey
-;; Keybindings don't overshadown comint bindings.
-;;
-;; Revision 1.22  1994/07/26  00:46:07  tromey
-;; Emacs 18 changes from Carl Witty.
-;;
-;; Revision 1.21  1994/07/14  22:49:21  tromey
-;; Added ";;;###autoload" comments where appropriate.
-;;
-; Revision 1.20  1994/06/05  16:57:22  tromey
-; tcl-current-word does the right thing in inferior-tcl-mode.
-;
-; Revision 1.19  1994/06/03  21:09:19  tromey
-; Another menu fix.
-;
-; Revision 1.18  1994/06/03  20:39:14  tromey
-; Fixed menu bug.
-;
-; Revision 1.17  1994/06/03  00:47:15  tromey
-; Fixed bug in bug-reporting code.
-;
-; Revision 1.16  1994/05/26  05:06:14  tromey
-; Menu items now sensitive as appropriate.
-;
-; Revision 1.15  1994/05/22  20:38:11  tromey
-; Added bug-report keybindings and menu entries.
-;
-; Revision 1.14  1994/05/22  20:18:28  tromey
-; Even more compile stuff.
-;
-; Revision 1.13  1994/05/22  20:17:15  tromey
-; Moved emacs version checking code to very beginning.
-;
-; Revision 1.12  1994/05/22  20:14:59  tromey
-; Compile fixes.
-;
-; Revision 1.11  1994/05/22  20:12:44  tromey
-; Fixed mark-defun for 19.23.
-; More menu fixes.
-;
-; Revision 1.10  1994/05/22  20:02:03  tromey
-; Fixed bug with M-;.
-; Wrote bug-reporting code.
-;
-; Revision 1.9  1994/05/22  05:26:51  tromey
-; Fixes for imenu.
-;
-; Revision 1.8  1994/05/22  03:38:07  tromey
-; Fixed menu support.
-;
-; Revision 1.7  1994/05/03  01:23:42  tromey
-; *** empty log message ***
-;
-; Revision 1.6  1994/04/23  16:23:36  tromey
-; Wrote tcl-indent-for-comment
-;
-;;
-;; 18-Mar-1994		Tom Tromey	Fourth beta release.
-;;    Added {un,}comment-region to menu.  Idea from
-;;    Mike Scheidler <c23mts@kocrsv01.delcoelect.com>
-;; 17-Mar-1994		Tom Tromey	
-;;    Fixed tcl-restart-with-file.  Bug fix attempt in
-;;    tcl-internal-end-of-defun.
-;; 16-Mar-1994		Tom Tromey	Third beta release
-;;    Added support code for menu (from Tcl mode written by
-;;    schmid@fb3-s7.math.TU-Berlin.DE (Gregor Schmid)).
-;; 12-Mar-1994		Tom Tromey	
-;;    Better documentation for inferior-tcl-buffer.  Wrote
-;;    tcl-restart-with-file.  Wrote Lucid Emacs menu (but no
-;;    code to install it).
-;; 12-Mar-1994		Tom Tromey	
-;;    Wrote tcl-guess-application.  Another stab at making
-;;    tcl-omit-ws-regexp work.
-;; 10-Mar-1994		Tom Tromey	Second beta release
-;;    Last Modified: Thu Mar 10 01:24:25 1994 (Tom Tromey)
-;;    Wrote perl-mode style line indentation command.
-;;    Wrote more documentation.  Added tcl-continued-indent-level.
-;;    Integrated help code.
-;; 8-Mar-1994		Tom Tromey	
-;;    Last Modified: Tue Mar  8 11:58:44 1994 (Tom Tromey)
-;;    Bug fixes.
-;; 6-Mar-1994		Tom Tromey	
-;;    Last Modified: Sun Mar  6 18:55:41 1994 (Tom Tromey)
-;;    Updated auto-newline support.
-;; 6-Mar-1994		Tom Tromey	Beta release
-;;    Last Modified: Sat Mar  5 17:24:32 1994 (Tom Tromey)
-;;    Wrote tcl-hashify-buffer.  Other minor bug fixes.
-;; 5-Mar-1994		Tom Tromey	
-;;    Last Modified: Sat Mar  5 16:11:20 1994 (Tom Tromey)
-;;    Wrote electric-hash code.
-;; 3-Mar-1994		Tom Tromey	
-;;    Last Modified: Thu Mar  3 02:53:40 1994 (Tom Tromey)
-;;    Added code to handle auto-fill in comments.
-;;    Added imenu support code.
-;;    Cleaned up code.
-;;    Better font-lock support.
-;; 28-Feb-1994		Tom Tromey	
-;;    Last Modified: Mon Feb 28 14:08:05 1994 (Tom Tromey)
-;;    Made tcl-figure-type more easily configurable.
-;; 28-Feb-1994		Tom Tromey	
-;;    Last Modified: Mon Feb 28 01:02:58 1994 (Tom Tromey)
-;;    Wrote inferior-tcl mode.
-;; 16-Feb-1994		Tom Tromey	
-;;    Last Modified: Wed Feb 16 17:05:19 1994 (Tom Tromey)
-;;    Added support for font-lock-mode.
-;; 29-Oct-1993		Tom Tromey	
-;;    Last Modified: Sun Oct 24 17:39:14 1993 (Tom Tromey)
-;;    Patches from Guido Bosch to make things work with Lucid Emacs.
-;; 22-Oct-1993		Tom Tromey	
-;;    Last Modified: Fri Oct 22 15:26:46 1993 (Tom Tromey)
-;;    Made many characters have "_" syntax class; suggested by Guido
-;;    Bosch <Guido.Bosch@loria.fr>.  Note that this includes the "$"
-;;    character, which might be a change you'd notice.
-;; 21-Oct-1993		Tom Tromey	
-;;    Last Modified: Thu Oct 21 20:28:40 1993 (Tom Tromey)
-;;    More fixes for tcl-omit-ws-regexp.
-;; 20-Oct-1993		Tom Tromey	
-;;    Started keeping history.  Fixed tcl-{beginning,end}-of-defun.
-;;    Added some code to make things work with Emacs 18.
-
-;; THANKS TO:
+;; THANKS FOR CRITICISM AND SUGGESTIONS TO:
 ;; Guido Bosch <Guido.Bosch@loria.fr>
 ;; pgs1002@esc.cam.ac.uk (Dr P.G. Sjoerdsma)
 ;; Mike Scheidler <c23mts@kocrsv01.delcoelect.com>
@@ -300,11 +64,9 @@
 ;; Jesper Pedersen <blackie@imada.ou.dk>
 ;; dfarmer@evolving.com (Doug Farmer)
 ;; "Chris Alfeld" <calfeld@math.utah.edu>
-;; Ben Wing <wing@666.com>
+;; Ben Wing <ben@xemacs.org>
 
 ;; KNOWN BUGS:
-;; * indent-region should skip blank lines.  (It does in v19, so I'm
-;;   not motivated to fix it here).
 ;; * In Tcl "#" is not always a comment character.  This can confuse
 ;;   tcl.el in certain circumstances.  For now the only workaround is
 ;;   to enclose offending hash characters in quotes or precede it with
@@ -312,10 +74,6 @@
 ;;   the syntax class of characters between them, while braces do not.
 ;;   The electric-# mode helps alleviate this problem somewhat.
 ;; * indent-tcl-exp is untested.
-;; * Doesn't work under Emacs 18 yet.
-;; * There's been a report that font-lock does strange things under
-;;   Lucid Emacs 19.6.  For instance in "proc foobar", the space
-;;   before "foobar" is highlighted.
 
 ;; TODO:
 ;; * make add-log-tcl-defun smarter.  should notice if we are in the
@@ -349,31 +107,37 @@
 
 ;;; Code:
 
+(eval-when-compile
+  (require 'imenu)
+  (require 'outline)
+  (require 'dabbrev)
+  (require 'add-log))
+
 ;; I sure wish Emacs had a package that made it easy to extract this
 ;; sort of information.  Strange definition works with XEmacs 20.0.
 (defconst tcl-using-emacs-19 (not (string-match "18\\." emacs-version))
-  "Nil unless using Emacs 19 (XEmacs or FSF).")
+  "Non-nil if using Emacs 19 or later.")
 
-;; FIXME this will break on Emacs 19.100.
 (defconst tcl-using-emacs-19-23
-  (string-match "19\\.\\(2[3-9]\\|[3-9][0-9]\\)" emacs-version)
-  "Nil unless using Emacs 19-23 or later.")
+  (or (string-match "19\\.\\(2[3-9]\\|[3-9][0-9]\\)" emacs-version)
+      (string-match "^[2-9][0-9]\\." emacs-version))
+  "Non-nil if using Emacs 19-23 or later.")
 
 (defconst tcl-using-xemacs-19 (string-match "XEmacs" emacs-version)
-  "Nil unless using XEmacs).")
+  "Non-nil if using XEmacs.")
 
 (require 'comint)
 
-;; When compiling under GNU Emacs, load imenu during compilation.  If
+;; When compiling under Emacs, load imenu during compilation.  If
 ;; you have 19.22 or earlier, comment this out, or get imenu.
 (and (fboundp 'eval-when-compile)
      (eval-when-compile
-       (if (and (string-match "19\\." emacs-version)
+       (if (and (not (string< emacs-version "19.23"))
 		(not (string-match "XEmacs" emacs-version)))
 	   (require 'imenu))
        ()))
 
-(defconst tcl-version "1.50")
+(defconst tcl-version "Revision: 1.64.4.1 ")
 (defconst tcl-maintainer "Tom Tromey <tromey@drip.colorado.edu>")
 
 ;;
@@ -381,25 +145,23 @@
 ;;
 
 (defgroup tcl nil
-  "Tcl programming language."
+  "Major mode for editing Tcl source in Emacs"
   :group 'languages)
 
-
 (defcustom tcl-indent-level 4
   "*Indentation of Tcl statements with respect to containing block."
-  :type 'integer
-  :group 'tcl)
+  :group 'tcl
+  :type 'integer)
 
 (defcustom tcl-continued-indent-level 4
   "*Indentation of continuation line relative to first line of command."
-  :type 'integer
-  :group 'tcl)
+  :group 'tcl
+  :type 'integer)
 
 (defcustom tcl-auto-newline nil
-  "*Non-nil means automatically newline before and after braces
-inserted in Tcl code."
-  :type 'boolean
-  :group 'tcl)
+  "*Non-nil means automatically newline before and after braces you insert."
+  :group 'tcl
+  :type 'boolean)
 
 (defcustom tcl-tab-always-indent t
   "*Control effect of TAB key.
@@ -415,49 +177,49 @@ to take place:
   4. Move forward to end of line, indenting if necessary.
   5. Create an empty comment.
   6. Move backward to start of comment, indenting if necessary."
-  :type '(choice (const :tag "on" t)
-		 (const :tag "off" nil)
-		 (sexp :format "%t\n" :tag "The Works" other))
-  :group 'tcl)
+  :group 'tcl
+  :type '(choice (const :tag "Always" t)
+		 (const :tag "Beginning only" nil)
+		 (const :tag "Maybe move or make or delete comment" 'tcl)))
+
 
 (defcustom tcl-use-hairy-comment-detector t
-  "*If not `nil', then the more complicated, but slower, comment
-detecting function is used.  This variable is only used in GNU Emacs
-19 (the fast function is always used elsewhere)."
-  :type 'boolean
-  :group 'tcl)
+  "*If not nil, use the more sophisticated, but slower, comment-delete method.
+This variable is not effective in Emacs 18;
+the fast function is always used in that version."
+  :group 'tcl
+  :type 'boolean)
 
 (defcustom tcl-electric-hash-style 'smart
   "*Style of electric hash insertion to use.
-Possible values are 'backslash, meaning that `\\' quoting should be
-done; 'quote, meaning that `\"' quoting should be done; 'smart,
-meaning that the choice between 'backslash and 'quote should be
+Possible values are `backslash', meaning that `\\' quoting should be
+done; `quote', meaning that `\"' quoting should be done; `smart',
+meaning that the choice between `backslash' and `quote' should be
 made depending on the number of hashes inserted; or nil, meaning that
 no quoting should be done.  Any other value for this variable is
-taken to mean 'smart.  The default is 'smart."
-  :type '(choice (const backslash) (const quote) (const smart))
-  :group 'tcl)
+taken to mean `smart'.  The default is `smart'."
+  :group 'tcl
+  :type '(choice (const backslash) (const quote) (const smart) (const nil)))
 
 (defcustom tcl-help-directory-list nil
-  "*List of topmost directories containing TclX help files"
-  :type '(repeat directory)
-  :group 'tcl)
+  "*List of topmost directories containing TclX help files."
+  :group 'tcl
+  :type '(repeat directory))
 
 (defcustom tcl-use-smart-word-finder t
-  "*If not nil, use a better way of finding the current word when
-looking up help on a Tcl command."
-  :type 'boolean
-  :group 'tcl)
+  "*If not nil, use smart way to find current word, for Tcl help feature."
+  :group 'tcl
+  :type 'boolean)
 
 (defcustom tcl-application "wish"
-  "*Name of Tcl application to run in inferior Tcl mode."
-  :type 'string
-  :group 'tcl)
+  "*Name of Tcl program to run in inferior Tcl mode."
+  :group 'tcl
+  :type 'string)
 
 (defcustom tcl-command-switches nil
-  "*Switches to supply to `tcl-application'."
-  :type '(repeat string)
-  :group 'tcl)
+  "*List of switches to supply to the `tcl-application' program."
+  :group 'tcl
+  :type '(repeat string))
 
 (defcustom tcl-prompt-regexp "^\\(% \\|\\)"
   "*If not nil, a regexp that will match the prompt in the inferior process.
@@ -465,8 +227,8 @@ If nil, the prompt is the name of the application with \">\" appended.
 
 The default is \"^\\(% \\|\\)\", which will match the default primary
 and secondary prompts for tclsh and wish."
-  :type 'regexp
-  :group 'tcl)
+  :group 'tcl
+  :type 'regexp)
 
 (defcustom inferior-tcl-source-command "source %s\n"
   "*Format-string for building a Tcl command to load a file.
@@ -474,8 +236,8 @@ This format string should use `%s' to substitute a file name
 and should result in a Tcl expression that will command the
 inferior Tcl to load that file.  The filename will be appropriately
 quoted for Tcl."
-  :type 'string
-  :group 'tcl)
+  :group 'tcl
+  :type 'string)
 
 ;;
 ;; Keymaps, abbrevs, syntax tables.
@@ -543,10 +305,10 @@ quoted for Tcl."
     ["Send bug report" tcl-submit-bug-report t])
   "XEmacs menu for Tcl mode.")
 
-;; GNU Emacs does menus via keymaps.  Do it in a function in case we
+;; Emacs does menus via keymaps.  Do it in a function in case we
 ;; later decide to add it to inferior Tcl mode as well.
-(defun tcl-add-fsf-menu (map)
-  (define-key map [menu-bar] (make-sparse-keymap))
+(defun tcl-add-emacs-menu (map)
+  (define-key map [menu-bar] (make-sparse-keymap "Tcl"))
   ;; This fails in Emacs 19.22 and earlier.
   (require 'lmenu)
   (let ((menu (make-lucid-menu-keymap "Tcl" tcl-xemacs-menu)))
@@ -569,6 +331,7 @@ quoted for Tcl."
   ;; FIXME.
   (define-key tcl-mode-map "\e\C-h" 'tcl-mark-defun)
   (define-key tcl-mode-map "\e\C-q" 'indent-tcl-exp)
+  (define-key tcl-mode-map "\177" 'backward-delete-char-untabify)
   (define-key tcl-mode-map "\t" 'tcl-indent-command)
   (define-key tcl-mode-map "\M-;" 'tcl-indent-for-comment)
   (define-key tcl-mode-map "\M-\C-x" 'tcl-eval-defun)
@@ -585,7 +348,7 @@ quoted for Tcl."
   ;; Make menus.
   (if (and tcl-using-emacs-19 (not tcl-using-xemacs-19))
       (progn
-	(tcl-add-fsf-menu tcl-mode-map))))
+	(tcl-add-emacs-menu tcl-mode-map))))
 
 (defun tcl-fill-inferior-map ()
   (define-key inferior-tcl-mode-map "\t" 'comint-dynamic-complete)
@@ -593,6 +356,7 @@ quoted for Tcl."
     'comint-dynamic-list-filename-completions)
   (define-key inferior-tcl-mode-map "\e\C-a" 'tcl-beginning-of-defun)
   (define-key inferior-tcl-mode-map "\e\C-e" 'tcl-end-of-defun)
+  (define-key inferior-tcl-mode-map "\177" 'backward-delete-char-untabify)
   (define-key inferior-tcl-mode-map "\M-\C-x" 'tcl-eval-defun)
   (define-key inferior-tcl-mode-map "\C-c\C-b" 'tcl-submit-bug-report)
   (define-key inferior-tcl-mode-map "\C-c\C-i" 'tcl-help-on-word)
@@ -662,21 +426,21 @@ Several functions exist which are useful to run from your
 `tcl-mode-hook' (see each function's documentation for more
 information):
 
-  tcl-guess-application
+  `tcl-guess-application'
     Guesses a default setting for `tcl-application' based on any
     \"#!\" line at the top of the file.
-  tcl-hashify-buffer
+  `tcl-hashify-buffer'
     Quotes all \"#\" characters that don't correspond to actual
     Tcl comments.  (Useful when editing code not originally created
     with this mode).
-  tcl-auto-fill-mode
+  `tcl-auto-fill-mode'
     Auto-filling of Tcl comments.
 
-Emacs 19 users can add functions to the hook with `add-hook':
+Add functions to the hook with `add-hook':
 
    (add-hook 'tcl-mode-hook 'tcl-guess-application)
 
-Emacs 18 users must use `setq':
+Emacs 18 users must use `setq' instead:
 
    (setq tcl-mode-hook (cons 'tcl-guess-application tcl-mode-hook))"
   :type 'hook
@@ -689,9 +453,9 @@ Emacs 18 users must use `setq':
   :group 'tcl)
 
 (defvar tcl-proc-list
-  '("proc" "method" "itcl_class")
+  '("proc" "method" "itcl_class" "body" "configbody" "class")
   "List of commands whose first argument defines something.
-This exists because some people (eg, me) use \"defvar\" et al.
+This exists because some people (eg, me) use `defvar' et al.
 Call `tcl-set-proc-regexp' and `tcl-set-font-lock-keywords'
 after changing this list.")
 
@@ -699,7 +463,8 @@ after changing this list.")
   "Regexp to use when matching proc headers.")
 
 (defvar tcl-typeword-list
-  '("global" "upvar" "inherit" "public" "protected" "common")
+  '("global" "upvar" "inherit" "public" "protected" "private"
+    "common" "itk_option" "variable")
   "List of Tcl keywords denoting \"type\".  Used only for highlighting.
 Call `tcl-set-font-lock-keywords' after changing this list.")
 
@@ -708,7 +473,7 @@ Call `tcl-set-font-lock-keywords' after changing this list.")
   '("if" "then" "else" "elseif" "for" "foreach" "break" "continue" "while"
     "eval" "case" "in" "switch" "default" "exit" "error" "proc" "return"
     "uplevel" "constructor" "destructor" "itcl_class" "loop" "for_array_keys"
-    "for_recursive_glob" "for_file")
+    "for_recursive_glob" "for_file" "method" "body" "configbody" "class")
   "List of Tcl keywords.  Used only for highlighting.
 Default list includes some TclX keywords.
 Call `tcl-set-font-lock-keywords' after changing this list.")
@@ -772,7 +537,6 @@ is a Tcl expression, and the last argument is Tcl commands.")
 ;; Work around differences between various versions of Emacs.
 ;;
 
-;; We use this because Lemacs 19.9 has what we need.
 (defconst tcl-pps-has-arg-6
   (or tcl-using-emacs-19
       (and tcl-using-xemacs-19
@@ -781,8 +545,7 @@ is a Tcl expression, and the last argument is Tcl commands.")
 		 (parse-partial-sexp (point) (point) nil nil nil t)
 		 t)
 	     (error nil))))
-  "t if using an emacs which supports sixth (\"commentstop\") argument
-to parse-partial-sexp.")
+  "t if Emacs supports COMMENTSTOP argument to `parse-partial-sexp'.")
 
 ;; Its pretty bogus to have to do this, but there is no easier way to
 ;; say "match not syntax-1 and not syntax-2".  Too bad you can't put
@@ -804,10 +567,11 @@ to parse-partial-sexp.")
 ;; problem seems to be that there is a bug in Emacs 19.22 where
 ;; end-of-defun doesn't really use the brace matching the one that
 ;; trails defun-prompt-regexp.
+;; ?? Is there a bug now ??
 (defconst tcl-omit-ws-regexp "^[^ \t\n#}][^\n}]+}*[ \t]+")
 
 (defun tcl-internal-beginning-of-defun (&optional arg)
-  "Move backward to next beginning-of-defun.
+  "Move backward to next beginning of defun.
 With argument, do this that many times.
 Returns t unless search stops due to end of buffer."
   (interactive "p")
@@ -847,18 +611,18 @@ An end of a defun is found by moving forward from the beginning of one."
       (forward-line)
       (if (> (point) start) (setq arg (1- arg))))))
 
-;; In Emacs 19, we can use begining-of-defun as long as we set up a
+;; We can now use begining-of-defun as long as we set up a
 ;; certain regexp.  In Emacs 18, we need our own function.
-(fset 'tcl-beginning-of-defun
-      (if tcl-using-emacs-19
-	  'beginning-of-defun
-	'tcl-internal-beginning-of-defun))
+(defalias 'tcl-beginning-of-defun
+  (if tcl-using-emacs-19
+      'beginning-of-defun
+    'tcl-internal-beginning-of-defun))
 
 ;; Ditto end-of-defun.
-(fset 'tcl-end-of-defun
-      (if (and tcl-using-emacs-19 (not tcl-using-xemacs-19))
-	  'end-of-defun
-	'tcl-internal-end-of-defun))
+(defalias 'tcl-end-of-defun
+  (if (and tcl-using-emacs-19 (not tcl-using-xemacs-19))
+      'end-of-defun
+    'tcl-internal-end-of-defun))
 
 ;; Internal mark-defun that is used for losing Emacsen.
 (defun tcl-internal-mark-defun ()
@@ -872,14 +636,14 @@ An end of a defun is found by moving forward from the beginning of one."
   (tcl-beginning-of-defun)
   (backward-paragraph))
 
-;; In GNU Emacs 19-23 and later, mark-defun works as advertised.  I
+;; In Emacs 19.23 and later, mark-defun works as advertised.  I
 ;; don't know about XEmacs, so for now it and Emacs 18 just lose.
 (fset 'tcl-mark-defun
       (if tcl-using-emacs-19-23
 	  'mark-defun
 	'tcl-internal-mark-defun))
 
-;; In GNU Emacs 19, mark takes an additional "force" argument.  I
+;; In Emacs 19, mark takes an additional "force" argument.  I
 ;; don't know about XEmacs, so I'm just assuming it is the same.
 ;; Emacs 18 doesn't have this argument.
 (defun tcl-mark ()
@@ -934,6 +698,10 @@ Uses variables `tcl-proc-regexp' and `tcl-keyword-list'."
     ()
   (tcl-set-font-lock-keywords))
 
+;; XEmacs change.
+(defvar tcl-imenu-generic-expression
+  '((nil "^proc[ \t]+\\([-A-Za-z0-9_:+*]+\\)" 1))
+  "Imenu generic expression for `tcl-mode'.  See `imenu-generic-expression'.")
 
 
 ;;
@@ -949,24 +717,24 @@ Paragraphs are separated by blank lines only.
 Delete converts tabs to spaces as it moves back.
 
 Variables controlling indentation style:
-  tcl-indent-level
+  `tcl-indent-level'
     Indentation of Tcl statements within surrounding block.
-  tcl-continued-indent-level
+  `tcl-continued-indent-level'
     Indentation of continuation line relative to first line of command.
 
 Variables controlling user interaction with mode (see variable
 documentation for details):
-  tcl-tab-always-indent
+  `tcl-tab-always-indent'
     Controls action of TAB key.
-  tcl-auto-newline
+  `tcl-auto-newline'
     Non-nil means automatically newline before and after braces, brackets,
     and semicolons inserted in Tcl code.
-  tcl-electric-hash-style
+  `tcl-electric-hash-style'
     Controls action of `#' key.
-  tcl-use-hairy-comment-detector
+  `tcl-use-hairy-comment-detector'
     If t, use more complicated, but slower, comment detector.
-    This variable is only used in GNU Emacs 19.
-  tcl-use-smart-word-finder
+    This variable is only used in Emacs 19.
+  `tcl-use-smart-word-finder'
     If not nil, use a smarter, Tcl-specific way to find the current
     word when looking up help on a Tcl command.
 
@@ -990,19 +758,24 @@ Commands:
   (if (and tcl-using-emacs-19-23
 	   (>= emacs-minor-version 29))
       (progn
-	;; In Emacs 19.29, you aren't supposed to start these with a
-	;; ^.
+	;; In Emacs 19.29, you aren't supposed to start these with a ^.
 	(setq paragraph-start "$\\|")
 	(setq paragraph-separate paragraph-start))
     (setq paragraph-start (concat "^$\\|" page-delimiter))
     (setq paragraph-separate paragraph-start))
   (make-local-variable 'paragraph-ignore-fill-prefix)
   (setq paragraph-ignore-fill-prefix t)
-  (make-local-variable 'fill-paragraph-function)
-  (setq fill-paragraph-function 'tcl-do-fill-paragraph)
+
+  ;; XEmacs change: check for filladapt-mode.
+  (if (not (and (boundp 'filladapt-mode)
+                filladapt-mode))
+      (progn
+        (make-local-variable 'fill-paragraph-function)
+        (setq fill-paragraph-function 'tcl-do-fill-paragraph)))
 
   (make-local-variable 'indent-line-function)
   (setq indent-line-function 'tcl-indent-line)
+
   ;; Tcl doesn't require a final newline.
   ;; (make-local-variable 'require-final-newline)
   ;; (setq require-final-newline t)
@@ -1021,12 +794,15 @@ Commands:
   (make-local-variable 'outline-level)
   (setq outline-level 'tcl-outline-level)
 
-  (make-local-variable 'font-lock-keywords)
-  (setq font-lock-keywords tcl-font-lock-keywords)
+  (make-local-variable 'font-lock-defaults)
+  (setq font-lock-defaults
+	'(tcl-font-lock-keywords))
+
+  ;; XEmacs change.
+  (make-local-variable 'imenu-generic-expression)
+  (setq imenu-generic-expression tcl-imenu-generic-expression)
+  ;; (add-hook 'tcl-mode-hook 'imenu-add-menubar-index)
 
-  ;; The following only really makes sense under GNU Emacs 19.
-  (make-local-variable 'imenu-create-index-function)
-  (setq imenu-create-index-function 'tcl-imenu-create-index-function)
   (make-local-variable 'parse-sexp-ignore-comments)
 
   ;; Settings for new dabbrev code.
@@ -1056,9 +832,8 @@ Commands:
     (setq parse-sexp-ignore-comments nil))
 
   ;; Put Tcl menu into menubar for XEmacs.  This happens
-  ;; automatically for GNU Emacs.
+  ;; automatically in Emacs.
   (if (and tcl-using-xemacs-19
-	   (featurep 'menubar)
 	   current-menubar
 	   (not (assoc "Tcl" current-menubar)))
       (progn
@@ -1120,10 +895,10 @@ Commands:
 
 (defun tcl-indent-command (&optional arg)
   "Indent current line as Tcl code, or in some cases insert a tab character.
-If tcl-tab-always-indent is t (the default), always indent current line.
-If tcl-tab-always-indent is nil and point is not in the indentation
+If `tcl-tab-always-indent' is t (the default), always indent current line.
+If `tcl-tab-always-indent' is nil and point is not in the indentation
 area at the beginning of the line, a TAB is inserted.
-Other values of tcl-tab-always-indent cause the first possible action
+Other values of `tcl-tab-always-indent' cause the first possible action
 from the following list to take place:
 
   1. Move from beginning of line to correct indentation.
@@ -1138,7 +913,7 @@ from the following list to take place:
     ;; Indent if in indentation area, otherwise insert TAB.
     (if (<= (current-column) (current-indentation))
 	(tcl-indent-line)
-      (self-insert-command arg)))
+      (insert-tab arg)))
    ((eq tcl-tab-always-indent t)
     ;; Always indent.
     (tcl-indent-line))
@@ -1220,7 +995,7 @@ Return the amount the indentation changed by."
 
 (defun tcl-figure-type ()
   "Determine type of sexp at point.
-This is either 'tcl-expr, 'tcl-commands, or nil.  Puts point at start
+This is either `tcl-expr', `tcl-commands', or nil.  Puts point at start
 of sexp that indicates types.
 
 See documentation for variable `tcl-type-alist' for more information."
@@ -1468,23 +1243,6 @@ Returns nil if line starts inside a string, t if in a comment."
 ;; Interfaces to other packages.
 ;;
 
-(defun tcl-imenu-create-index-function ()
-  "Generate alist of indices for imenu."
-  (let ((re (concat tcl-proc-regexp "\\([^ \t\n{]+\\)"))
-	alist prev-pos)
-    (goto-char (point-min))
-    (imenu-progress-message prev-pos 0)
-    (save-match-data
-      (while (re-search-forward re nil t)
-	(imenu-progress-message prev-pos)
-	;; Position on start of proc name, not beginning of line.
-	(setq alist (cons
-		     (cons (buffer-substring (match-beginning 2) (match-end 2))
-			   (match-beginning 2))
-		     alist))))
-    (imenu-progress-message prev-pos 100)
-    (nreverse alist)))
-
 ;; FIXME Definition of function is very ad-hoc.  Should use
 ;; tcl-beginning-of-defun.  Also has incestuous knowledge about the
 ;; format of tcl-proc-regexp.
@@ -1595,24 +1353,24 @@ Prefix argument means switch to the Tcl buffer afterwards."
 (defun inferior-tcl-mode ()
   "Major mode for interacting with Tcl interpreter.
 
-A Tcl process can be started with M-x inferior-tcl.
+A Tcl process can be started with \\[inferior-tcl].
 
-Entry to this mode runs the hooks comint-mode-hook and
-inferior-tcl-mode-hook, in that order.
+Entry to this mode runs the normal hooks `comint-mode-hook' and
+`inferior-tcl-mode-hook', in that order.
 
 You can send text to the inferior Tcl process from other buffers
 containing Tcl source.
 
 Variables controlling Inferior Tcl mode:
-  tcl-application
+  `tcl-application'
     Name of program to run.
-  tcl-command-switches
+  `tcl-command-switches'
     Command line arguments to `tcl-application'.
-  tcl-prompt-regexp
+  `tcl-prompt-regexp'
     Matches prompt.
-  inferior-tcl-source-command
+  `inferior-tcl-source-command'
     Command to use to read Tcl file in running application.
-  inferior-tcl-buffer
+  `inferior-tcl-buffer'
     The current inferior Tcl process buffer.  See variable
     documentation for details on multiple-process support.
 
@@ -1701,7 +1459,9 @@ semicolon, opening brace, or opening bracket on the same line."
 Returns a list of the form `(FLAG . STATE)'.  STATE can be used
 as input to future invocations.  FLAG is nil if not in comment,
 t otherwise.  If in comment, leaves point at beginning of comment.
-Only works in Emacs 19.  See also `tcl-simple-scan-for-comment', a
+
+This function does not work in Emacs 18.
+See also `tcl-simple-scan-for-comment', a
 simpler version that is often right, and works in Emacs 18."
   (let ((bol (save-excursion
 	       (goto-char end)
@@ -1740,16 +1500,14 @@ simpler version that is often right, and works in Emacs 18."
     (cons real-comment state)))
 
 (defun tcl-hairy-in-comment ()
-  "Return t if point is in a comment, and leave point at beginning
-of comment."
+  "Return t if point is in a comment, and leave point at beginning of comment."
   (let ((save (point)))
     (tcl-beginning-of-defun)
     (car (tcl-hairy-scan-for-comment nil save nil))))
 
 (defun tcl-simple-in-comment ()
-  "Return t if point is in comment, and leave point at beginning
-of comment.  This is faster that `tcl-hairy-in-comment', but is
-correct less often."
+  "Return t if point is in comment, and leave point at beginning of comment.
+This is faster that `tcl-hairy-in-comment', but is correct less often."
   (let ((save (point))
 	comment)
     (beginning-of-line)
@@ -1759,15 +1517,14 @@ correct less often."
     comment))
 
 (defun tcl-in-comment ()
-  "Return t if point is in comment, and leave point at beginning
-of comment."
+  "Return t if point is in comment, and leave point at beginning of comment."
   (if (and tcl-pps-has-arg-6
 	   tcl-use-hairy-comment-detector)
       (tcl-hairy-in-comment)
     (tcl-simple-in-comment)))
 
 (defun tcl-do-fill-paragraph (ignore)
-  "fill-paragraph function for Tcl mode.  Only fills in a comment."
+  "`fill-paragraph' function for Tcl mode.  Only fills in a comment."
   (let (in-comment col where)
     (save-excursion
       (end-of-line)
@@ -1872,7 +1629,7 @@ to update the alist.")
   (message "Building Tcl help file index...done"))
 
 (defun tcl-word-no-props ()
-  "Like current-word, but strips properties."
+  "Like `current-word', but strips properties."
   (let ((word (current-word)))
     (and (fboundp 'set-text-properties)
 	 (set-text-properties 0 (length word) nil word))
@@ -2008,13 +1765,11 @@ Prefix argument means switch to the Tcl buffer afterwards."
 		     tcl-application file tcl-command-switches)
 	(if and-go (switch-to-tcl t)))))))
 
-;; FIXME I imagine you can do this under Emacs 18.  I just don't know
-;; how.
 (defun tcl-auto-fill-mode (&optional arg)
   "Like `auto-fill-mode', but controls filling of Tcl comments."
   (interactive "P")
   (and (not tcl-using-emacs-19)
-       (error "You must use Emacs 19 to get this feature."))
+       (error "This feature is not supported in Emacs 18"))
   ;; Following code taken from "auto-fill-mode" (simple.el).
   (prog1
       (setq auto-fill-function
@@ -2107,7 +1862,7 @@ styles."
 (defun tcl-indent-for-comment ()
   "Indent this line's comment to comment column, or insert an empty comment.
 Is smart about syntax of Tcl comments.
-Parts of this were taken from indent-for-comment (simple.el)."
+Parts of this were taken from `indent-for-comment'."
   (interactive "*")
   (end-of-line)
   (or (tcl-in-comment)
@@ -2127,9 +1882,7 @@ Parts of this were taken from indent-for-comment (simple.el)."
 		(backward-char))))))
   ;; Point is just after the "#" starting a comment.  Move it as
   ;; appropriate.
-  (let* ((indent (if comment-indent-hook
-		     (funcall comment-indent-hook)
-		   (funcall comment-indent-function)))
+  (let* ((indent (or (funcall comment-indent-function) comment-column))
 	 (begpos (progn
 		   (backward-char)
 		   (point))))
@@ -2180,7 +1933,7 @@ The first line is assumed to look like \"#!.../program ...\"."
 ;; XEmacs menu support.
 ;; Taken from schmid@fb3-s7.math.TU-Berlin.DE (Gregor Schmid),
 ;; who wrote a different Tcl mode.
-;; We also have support for menus in FSF.  We do this by
+;; We also have support for menus in Emacs.  We do this by
 ;; loading the XEmacs menu emulation code.
 ;;
 
@@ -2197,7 +1950,7 @@ The first line is assumed to look like \"#!.../program ...\"."
 	 ;; Both of these problems are fixed in Emacs 19.23.  People
 	 ;; using an Emacs before that just suffer.
 	 (require 'menubar "lmenu")))  ;; This is annoying
-  ;; IMHO popup-menu should be autoloaded in FSF Emacs.  Oh well.
+  ;; IMHO popup-menu should be autoloaded.  Oh well.
   (popup-menu tcl-xemacs-menu))
 
 
@@ -2222,6 +1975,10 @@ The first line is assumed to look like \"#!.../program ...\"."
 ;; Bug reporting.
 ;;
 
+(and (fboundp 'eval-when-compile)
+     (eval-when-compile
+       (require 'reporter)))
+
 (defun tcl-submit-bug-report ()
   "Submit via mail a bug report on Tcl mode."
   (interactive)
@@ -2254,6 +2011,12 @@ The first line is assumed to look like \"#!.../program ...\"."
       tcl-pps-has-arg-6))))
 
 
+;; XEmacs additions
+;;;###autoload
+(add-to-list 'auto-mode-alist '("\\.\\(?:tcl\\|exp\\)\\'" . tcl-mode))
+;;;###autoload
+(add-to-list 'interpreter-mode-alist
+             '("^#!.*\\b\\(scope\\|wish\\|tcl\\|tclsh\\|expect\\)" . tcl-mode))
 
 (provide 'tcl)
 
